@@ -7,7 +7,6 @@ from copy import deepcopy
 from tqdm.autonotebook import tqdm
 from multiprocessing import Process, Manager, Queue, Pool, Array
 from time import sleep
-import psutil
 
 from medcat.cdb import CDB
 from medcat.preprocessing.tokenizers import spacy_split_all
@@ -720,13 +719,12 @@ class CAT(object):
         return json.dumps(out)
 
 
-    def multiprocessing(self, in_data, nproc=8, batch_size_chars=1000000, max_ram_percentage=-1, only_cui=False, addl_info=[]):
+    def multiprocessing(self, in_data, nproc=8, batch_size_chars=1000000, only_cui=False, addl_info=[]):
         r''' Run multiprocessing NOT FOR TRAINING
 
         in_data:  an iterator or array with format: [(id, text), (id, text), ...]
         nproc:  number of processors
         batch_size_chars: size of a batch in number of characters
-        max_ram_percentage: maximum utilzation of RAM when running many processes, if it is more, it will pause until some processes are done 
 
         return:  an list of tuples: [(id, doc_json), (id, doc_json), ...]
         '''
@@ -745,7 +743,7 @@ class CAT(object):
         procs = []
         for i in range(nproc):
             p = Process(target=self._mp_cons, kwargs={'in_q': in_q, 'out_dict': out_dict, 'pid': i, 'only_cui': only_cui,
-                'addl_info': addl_info, 'max_ram_percentage': max_ram_percentage})
+                'addl_info': addl_info})
             p.start()
             procs.append(p)
 
@@ -784,7 +782,7 @@ class CAT(object):
         return out
 
 
-    def _mp_cons(self, in_q, out_dict, pid=0, only_cui=False, addl_info=[], max_ram_percentage=-1):
+    def _mp_cons(self, in_q, out_dict, pid=0, only_cui=False, addl_info=[]):
         cnt = 0
         out = []
         while True:
@@ -796,11 +794,6 @@ class CAT(object):
 
                 for id, text in data:
                     try:
-                        if max_ram_percentage > 0:
-                            while psutil.virtual_memory().percent > max_ram_percentage:
-                                self.log.debug("Process: {}, waiting because current RAM usage: {}".format(pid, psutil.virtual_memory().percent)
-                                sleep(30)
-
                         # Annotate document
                         doc = self.get_entities(text=text, only_cui=only_cui, addl_info=addl_info)
                         doc['text'] = text
